﻿// ********************************************************************************
// SlidePrinter.cs
// 04/2012
// Primera Technology, Inc.
// ********************************************************************************

using System;
using System.Collections;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;

namespace PrintDriverIntegration
{
    public enum SPCoverFlag
    {
        Unknown, Closed, Open
    }
    
    public enum SPPrintingStatus
    {
        PrinterReady=0, PrinterBusy=0x80
    }

    public enum SPSystemErrors
    {
        HeadLifterError, InvalidRibbon, RibbonAdvanceError, PaperJamError, PaperOut, HeadLoadingError, RibbonPWMError, InvalidCartridge, RibbonOut, RibbonBreak, ModelNumber, OutputFull, SlideCartridgeContact, RibbonSupplyContact, PrinterError = 98,None =99
    }

    public enum SPSystemStatus
    {
        Idle, Busy, PowerOff
    }

    public enum SPRibbonTypes
    {
        Unknown, Mono, Cyan, Magenta, Yellow, Black
    }

    public enum RibbonModels
    {
        Invalid=0, Mono, TwoColor, ThreeColor, CMYK, Missing=0xff
    }

    public enum SlideModels
    {
        Invalid = 0, Valid, Missing = 0xff
    }

    class SlidePrinter
    {

// ReSharper disable InconsistentNaming
        private static readonly SlidePrinter instance = new SlidePrinter();
// ReSharper restore InconsistentNaming

        private int _slidePercentRemaining;
        private int _ribbonPercentRemaining;
        private bool LastStatusValid { get; set; }
        private string Printer { get; set; }
        private readonly PrintLabel _printLabel;
        /// <summary>
        /// Public Properties
        ///     The structure returned from the Slide Printer is completely defined in the Integration
        ///     Guide.  We have included here a subset of the most used status values returned by the slide
        ///     printer.
        /// 
        /// </summary>
        /// 

        public const byte SlideMediaMissing = 0xff;
        public const byte SlideMediaInvalid = 0xfe;
        public bool InPreloadMode { get; private set; }
        public byte RecordNum { get; private set; }
        public byte Model { get; private set; }
        public byte OEMModel { get; private set; }
        public byte PGAVersion { get; private set; }
        public string FWVersion { get; private set; }
        public string FWChksum { get; private set; }
        public SPPrintingStatus PrintingStatus { get; private set; }
        public ArrayList SystemError { get; private set; }
        public SPSystemStatus SystemBusy { get; private set; }
        public SPCoverFlag CoverFlag { get; private set; }
        public byte MediaType { get; private set; }
        public RibbonModels RibbonType { get; private set; }
        public UInt16 RibbonRemaining { get; private set; }
        public string SlideCartridgeSerial { get; private set; }
        public string RibbonSerial { get; private set; }
        public UInt16 SlidesRemaining { get; private set; }
        public UInt32 SlidesPrinted { get; private set; }
        public UInt16 ErrorBytes { get; private set; }
        public bool SlideOutputFull { get; private set; }
        public bool CartridgeHasBeenRemoved { get; private set; }
        public byte SlideTypeMonoAdjust { get; set; }
        public byte SlideTypeColorAdjust { get; set; }
        public byte TOFOffset { get; private set; }
        public byte HorzOffset { get; private set; }
        public byte OverallColorPrintHeat { get; private set; }
        public byte OverallBlackPrintHeat { get; private set; }

        

        private SlidePrinter() 
        {
            LastStatusValid = false;
            FWVersion = "?";
            FWChksum = "?";
            SystemError = new ArrayList();
            _printLabel = new PrintLabel();
        }

        public static SlidePrinter Instance
        {
            get
            {
                return instance;
            }
        }

        public int RibbonPercentRemaining
        {
            get
            {
                if (RibbonType == RibbonModels.Mono)
                    _ribbonPercentRemaining = (int)(((double)RibbonRemaining / 5000) * 100);
                else
                    _ribbonPercentRemaining = (int)(((double)RibbonRemaining / 1000) * 100);

                return _ribbonPercentRemaining;
            }
            set
            {
                _ribbonPercentRemaining = value;
            }
        }

        public bool PrintImageFile(string file)
        {
            _printLabel.File = file;
            try
            {
                _printLabel.PrintImageFile(87, 67);
            }
            catch (Exception)
            {
                return false;
            }

            return true;
        }


        public int SlidePercentRemaining
        {
            get
            {
                _slidePercentRemaining = (int)(((double)SlidesRemaining / 100) * 100);
                return _slidePercentRemaining;
            }
            set
            {
                _slidePercentRemaining = value;
            }
        }

        public void SetPrinter(string printer)
        {
            Printer = printer;
            _printLabel.Printer = printer;
        }

        private bool GetFlashChipInfo(bool ribbon, ref string manf, ref String description, ref String lotNumber)
        {
            lock (this)
            {
                if (ribbon)
                {
                    try
                    {
                        byte[] cmd = new byte[8];
                        cmd[0] = 0x1b;
                        cmd[1] = 0x04;
                        cmd[2] = 0x12;

                        cmd[3] = 0x41;
                        cmd[4] = 0;
                        cmd[5] = 0;
                        cmd[6] = 0;

                        byte bChkSum = 0;
                        for (int i = 0; i < cmd.Length - 1; i++)
                            bChkSum += cmd[i];
                        cmd[7] = bChkSum;

                        // Your unmanaged pointer.

                        // Allocate some unmanaged memory for those bytes.
                        IntPtr pUnmanagedBytes = Marshal.AllocCoTaskMem(8);
                        // Copy the managed byte array into the unmanaged array.
                        Marshal.Copy(cmd, 0, pUnmanagedBytes, 8);
                        // Send the unmanaged bytes to the printer.

                        //_USBprinter.WriteData(_USBHandle, pUnmanagedBytes, 8);

                        if (_printLabel != null)
                            _printLabel.SendBytesToPrinter(_printLabel.Printer, pUnmanagedBytes, 8);
                    }
                    catch (Exception)
                    {
                        return false;
                    }
                }

                try
                {
                    byte[] bytes = new byte[263];
                    Int32 nBytes = 0;

                    if (_printLabel != null && _printLabel.ReadBytesFromPrinter(_printLabel.Printer, bytes, 263, ref nBytes, "PRIMERA"))
                    {
                        if (nBytes == 263)
                        {
                            String comments = Encoding.ASCII.GetString(bytes, 84, 34).TrimEnd();
                            String[] splitStr = comments.Split(';');

                            if (splitStr.Length == 2)
                            {
                                description = splitStr[0];
                            }
                            else if (splitStr.Length == 3)
                            {
                                description = splitStr[0];
                                lotNumber = splitStr[2];
                            }
                            else
                            {
                                String[] newSplitStr = comments.Split(' ');
                                if (newSplitStr.Length > 0 && newSplitStr[newSplitStr.Length - 1].Length == 2)
                                {
                                    description = comments.Remove(comments.Length - 2);
                                }
                                else
                                    description = comments;
                            }

                            manf = Encoding.ASCII.GetString(bytes, 118, 34).TrimEnd();

                        }
                    }
                }
                catch (Exception)
                {
                    return false;
                }
            }
            return true;
        }

        public override string ToString()
        {
            string sRet = "";
            const string sBrk = "\r\n";

            sRet += "----  Slide Printer ----" + sBrk;

            if (!LastStatusValid)
            {
                sRet += "    NOT CONNECTED" + sBrk;
            }
            else
            {

                sRet += "    Record: " + RecordNum + sBrk;
                sRet += "    FW Ver: " + FWVersion + "(" + FWChksum + ")" + sBrk;
                sRet += "    Status: " + SystemBusy.ToString() + sBrk;
                sRet += "    Printing Status: " + PrintingStatus.ToString() + sBrk;
                sRet += "    Cover: " + CoverFlag.ToString() + sBrk;

                sRet += "    Slide Output Full: " + SlideOutputFull.ToString() + sBrk;
                sRet += "    Cartridge Has Been Removed: " + CartridgeHasBeenRemoved.ToString() + sBrk;
                if (SystemError.Count > 0)
                {
                    foreach(SPSystemErrors err in SystemError)
                        sRet += "    Error: " + err.ToString() + sBrk;
                }
                else
                    sRet += "    Error: None" + sBrk;

                
                sRet += sBrk +"    Ribbon Type: " + RibbonType.ToString() + sBrk;
                try
                {
                    sRet += "    Ribbon Serial: " + RibbonSerial + sBrk;
                }
                catch
                {
                    sRet += "    Ribbon Serial: ??" + sBrk;
                }
                sRet += "    Ribbons Remaining: " + RibbonRemaining.ToString() + sBrk;



                sRet += sBrk + "    Slide Cartridge Type: " + MediaType.ToString() + sBrk;
                try
                {
                    sRet += "    Slide Cartridge Serial: " + SlideCartridgeSerial + sBrk;
                }
                catch 
                {
                    sRet += "    Slide Cartridge Serial: ??" + sBrk;
                }
                sRet += "    Slides Remaining: " + SlidesRemaining.ToString() + sBrk;
                sRet += "    Slides Printed: " + SlidesPrinted.ToString() + sBrk;

                sRet += sBrk + "    TOF Offset: " + TOFOffset.ToString() + sBrk;
                sRet += "    Horz Offset: " + HorzOffset.ToString() + sBrk;
                sRet += "    Overall Color Print Heat: " + OverallColorPrintHeat.ToString() + sBrk;
                sRet += "    Overall Black Print Heat: " + OverallBlackPrintHeat.ToString() + sBrk;

                
            }
            return sRet;
        }

        private void ProcessErrorByte(UInt16 bError)
        {
            SystemError.Clear();
            
            if ((bError & 0x1) > 0 )
                SystemError.Add(SPSystemErrors.HeadLifterError);
            if ((bError & 0x2) > 0)
                SystemError.Add(SPSystemErrors.InvalidRibbon);
            if ((bError & 0x4) > 0)
                SystemError.Add(SPSystemErrors.RibbonAdvanceError);
            if ((bError & 0x8) > 0)
                SystemError.Add(SPSystemErrors.PaperJamError);
            if ((bError & 0x10) > 0)
                SystemError.Add(SPSystemErrors.PaperOut);
            if ((bError & 0x20) > 0)
                SystemError.Add(SPSystemErrors.HeadLoadingError);
            if ((bError & 0x40) > 0)
                SystemError.Add(SPSystemErrors.RibbonPWMError);
            if ((bError & 0x80) > 0)
                SystemError.Add(SPSystemErrors.InvalidCartridge);
            if ((bError & 0x100) > 0)
                SystemError.Add(SPSystemErrors.RibbonOut);
            if ((bError & 0x200) > 0)
                SystemError.Add(SPSystemErrors.RibbonBreak);
            if ((bError & 0x400) > 0)
                SystemError.Add(SPSystemErrors.ModelNumber);
            if ((bError & 0x800) > 0)
                SystemError.Add(SPSystemErrors.OutputFull);
            if ((bError & 0x1000) > 0)
                SystemError.Add(SPSystemErrors.SlideCartridgeContact);
            if ((bError & 0x2000) > 0)
                SystemError.Add(SPSystemErrors.RibbonSupplyContact);
        }

        public bool PreloadSlide()
        {
            bool bRet = true;

            if (Refresh(2, 30))
            {
                try
                {
                    byte[] cmd = new byte[24];

                    for (int iInit = 0; iInit < cmd.Length; iInit++)
                    {
                        cmd[iInit] = 0x00;
                    }

                    cmd[0] = 0x1b;
                    cmd[1] = 0x51;
                    cmd[2] = 0x16;

                    cmd[3] = 1;


                    if (RibbonType == RibbonModels.Mono)
                    {
                        cmd[4] = 1;
                        cmd[5] = OverallBlackPrintHeat;
                    }
                    else if (RibbonType == RibbonModels.CMYK)
                    {
                        cmd[4] = 4;
                        cmd[5] = OverallColorPrintHeat;
                    }
                    else
                    {
                        bRet = false;
                    }

                    if (bRet)
                    {
                        cmd[6] = 0;

                        cmd[7] = (byte)DateTime.Now.Month;
                        cmd[8] = (byte)DateTime.Now.Day;
                        cmd[9] = Convert.ToByte(DateTime.Now.ToString("yy"));



                        byte bChkSum = 0;
                        for (int i = 0; i < cmd.Length - 1; i++)
                            bChkSum += cmd[i];
                        cmd[23] = bChkSum;

                        // Allocate some unmanaged memory for those bytes.
                        IntPtr pUnmanagedBytes = Marshal.AllocCoTaskMem(24);
                        // Copy the managed byte array into the unmanaged array.
                        Marshal.Copy(cmd, 0, pUnmanagedBytes, 24);
                        // Send the unmanaged bytes to the printer.

                        //_USBprinter.WriteData(_USBHandle, pUnmanagedBytes, 8);

                        bRet = _printLabel.SendBytesToPrinter(_printLabel.Printer, pUnmanagedBytes, 24);
                    }
                }
                catch (Exception)
                {
                    bRet = false;
                }




            }
            else
            {
                bRet = false;
            }

            return bRet;

        }


        public bool Refresh(UInt32 nRetries, int timeBtwRetries)
        {
            lock (this)
            {
                while (true)
                {
                    try
                    {
                        byte[] bytes = new byte[128];
                        Int32 nBytes = 0;

                        if (_printLabel.ReadBytesFromPrinter(_printLabel.Printer, bytes, 128, ref nBytes, "PRIMERA"))
                        {
                            if (bytes[0] == 0x1b && nBytes == 128)
                            {
                                LastStatusValid = true;
                                RecordNum = bytes[5];
                                Model = bytes[6];
                                PGAVersion = bytes[7];
                                FWVersion = Encoding.ASCII.GetString(bytes, 8, 16);
                                FWChksum = Encoding.ASCII.GetString(bytes, 24, 4);
                                OEMModel = bytes[32];
                                ErrorBytes = (UInt16)((bytes[28]) + (UInt16)(bytes[29] << 8));
                                ProcessErrorByte((UInt16)((bytes[28]) + (UInt16)(bytes[29] << 8)));

                                if (bytes[75] == 'B')
                                    SystemBusy = SPSystemStatus.Busy;
                                else if (bytes[75] == 'X')
                                    SystemBusy = SPSystemStatus.PowerOff;
                                else
                                    SystemBusy = SPSystemStatus.Idle;

                                if (bytes[79] == 0x00)
                                    PrintingStatus = SPPrintingStatus.PrinterReady;
                                else
                                    PrintingStatus = SPPrintingStatus.PrinterBusy;


                                if (bytes[76] == 'C')
                                    CoverFlag = SPCoverFlag.Closed;
                                else if (bytes[76] == 'O')
                                    CoverFlag = SPCoverFlag.Open;
                                else
                                    CoverFlag = SPCoverFlag.Unknown;

                                MediaType = bytes[58];

                                if (bytes[59] == 0x1)
                                    RibbonType = RibbonModels.Mono;
                                else if (bytes[59] == 0x2)
                                    RibbonType = RibbonModels.TwoColor;
                                else if (bytes[59] == 0x3)
                                    RibbonType = RibbonModels.ThreeColor;
                                else if (bytes[59] == 0x4)
                                    RibbonType = RibbonModels.CMYK;
                                else if (bytes[59] == 0xff)
                                    RibbonType = RibbonModels.Missing;
                                else
                                    RibbonType = RibbonModels.Invalid;
                                //MediaType = bytes[58];
                                //RibbonType = bytes[59];

                                if (bytes[33] == 0x80)
                                    SlideOutputFull = true;
                                else
                                    SlideOutputFull = false;

                                SlideCartridgeSerial = "";
                                RibbonSerial = "";
                                int tmp;
                                for (tmp = 0; tmp < 8; tmp++)
                                    SlideCartridgeSerial += (bytes[61 + tmp]).ToString("X2");
                                for (tmp = 0; tmp < 8; tmp++)
                                    RibbonSerial += (bytes[100 + tmp]).ToString("X2");

                                SlidesRemaining = (UInt16)((bytes[70]) + (UInt16)(bytes[69] << 8));
                                RibbonRemaining = (UInt16)((bytes[109]) + (UInt16)(bytes[108] << 8));
                                SlidesPrinted = (bytes[74] + (UInt32)(bytes[73] << 8) + (UInt32)(bytes[72] << 16) + (UInt32)(bytes[71] << 24));

                                if (bytes[57] == 1)
                                    InPreloadMode = true;
                                else
                                    InPreloadMode = false;

                                TOFOffset = bytes[96];
                                HorzOffset = bytes[97];

                                OverallBlackPrintHeat = bytes[98];
                                OverallColorPrintHeat = bytes[99];

                                if (bytes[110] == 0)
                                    CartridgeHasBeenRemoved = false;
                                else
                                    CartridgeHasBeenRemoved = true;

                                break;
                            }
                        }
                    }
                    catch (Exception)
                    {
                    }

                    if (nRetries-- <= 0)
                    {
                        //trace.LogError( "SlidePrinter Status: Failed to get status (retries exceeded)");
                        LastStatusValid = false;
                        return false;
                    }

                    Thread.Sleep(timeBtwRetries);
                }
            }

            return true;
        }

        public bool SetOverallBlackPrintHeat(byte bValue)
        {
            return SendSystemSetupCmd(1, 12, bValue, 0);
        }

        public bool SetOverallColorPrintHeat(byte bValue)
        {
            return SendSystemSetupCmd(1, 13, bValue, 0);
        }

        public bool SetVertOffset(byte bValue)
        {
            return SendSystemSetupCmd(1, 10, bValue, 0);
        }
        public bool SetHorzOffset(byte bValue)
        {
            return SendSystemSetupCmd(1, 11, bValue, 0);
        }

        public bool SetSlideType(byte bValue)
        {
            bool bRet;
            try
            {
                byte[] cmd = new byte[8];
                cmd[0] = 0x1b;
                cmd[1] = 0x04;
                cmd[2] = 0x0b;

                cmd[3] = 0;
                cmd[4] = bValue;
                cmd[5] = 0;
                cmd[6] = 0;

                byte bChkSum = 0;
                for (int i = 0; i < cmd.Length - 1; i++)
                    bChkSum += cmd[i];
                cmd[7] = bChkSum;

                // Allocate some unmanaged memory for those bytes.
                IntPtr pUnmanagedBytes = Marshal.AllocCoTaskMem(8);
                // Copy the managed byte array into the unmanaged array.
                Marshal.Copy(cmd, 0, pUnmanagedBytes, 8);
                // Send the unmanaged bytes to the printer.

                bRet = _printLabel.SendBytesToPrinter(_printLabel.Printer, pUnmanagedBytes, 8);
            }
            catch (Exception)
            {
                bRet = false;
            }
            return bRet;
        }

        public bool SetSlideCount(byte bValue)
        {
            bool bRet;
            try
            {
                byte[] cmd = new byte[8];
                cmd[0] = 0x1b;
                cmd[1] = 0x04;
                cmd[2] = 0x0b;

                cmd[3] = 0x01;
                cmd[4] = 0;
                cmd[5] = bValue;
                cmd[6] = 0;

                byte bChkSum = 0;
                for (int i = 0; i < cmd.Length - 1; i++)
                    bChkSum += cmd[i];
                cmd[7] = bChkSum;

                // Your unmanaged pointer.

                // Allocate some unmanaged memory for those bytes.
                IntPtr pUnmanagedBytes = Marshal.AllocCoTaskMem(8);
                // Copy the managed byte array into the unmanaged array.
                Marshal.Copy(cmd, 0, pUnmanagedBytes, 8);
                // Send the unmanaged bytes to the printer.

                //_USBprinter.WriteData(_USBHandle, pUnmanagedBytes, 8);

                bRet = _printLabel.SendBytesToPrinter(_printLabel.Printer, pUnmanagedBytes, 8);
            }
            catch (Exception)
            {
                bRet = false;
            }
            return bRet;
        }

        public bool RestoreFacDefaults()
        {

            bool bRet;
            try
            {
                byte[] cmd = new byte[8];
                cmd[0] = 0x1b;
                cmd[1] = 0x04;
                cmd[2] = 0x15;

                cmd[3] = 0;
                cmd[4] = 0;
                cmd[5] = 0;
                cmd[6] = 0;

                byte bChkSum = 0;
                for (int i = 0; i < cmd.Length - 1; i++)
                    bChkSum += cmd[i];
                cmd[7] = bChkSum;

                // Your unmanaged pointer.

                // Allocate some unmanaged memory for those bytes.
                IntPtr pUnmanagedBytes = Marshal.AllocCoTaskMem(8);
                // Copy the managed byte array into the unmanaged array.
                Marshal.Copy(cmd, 0, pUnmanagedBytes, 8);
                // Send the unmanaged bytes to the printer.

                //_USBprinter.WriteData(_USBHandle, pUnmanagedBytes, 8);

                bRet = _printLabel.SendBytesToPrinter(_printLabel.Printer, pUnmanagedBytes, 8);
            }
            catch (Exception)
            {
                bRet = false;
            }
            return bRet;


        }

        public bool SendReset()
        {

            bool bRet;
            try
            {
                byte[] cmd = new byte[8];
                cmd[0] = 0x1b;
                cmd[1] = 0x04;
                cmd[2] = 0x05;

                cmd[3] = 0;
                cmd[4] = 0;
                cmd[5] = 0;
                cmd[6] = 0;

                byte bChkSum = 0;
                for (int i = 0; i < cmd.Length - 1; i++)
                    bChkSum += cmd[i];
                cmd[7] = bChkSum;

                // Your unmanaged pointer.

                // Allocate some unmanaged memory for those bytes.
                IntPtr pUnmanagedBytes = Marshal.AllocCoTaskMem(8);
                // Copy the managed byte array into the unmanaged array.
                Marshal.Copy(cmd, 0, pUnmanagedBytes, 8);
                // Send the unmanaged bytes to the printer.

                //_USBprinter.WriteData(_USBHandle, pUnmanagedBytes, 8);

                bRet = _printLabel.SendBytesToPrinter(_printLabel.Printer, pUnmanagedBytes, 8);
            }
            catch (Exception)
            {
                bRet = false;
            }
            return bRet;


        }

        private bool SendSystemSetupCmd(byte nBytes, byte offset, byte b1, byte b2)
        {
            
            bool bRet;
            try
            {
                byte[] cmd = new byte[8];
                cmd[0] = 0x1b;
                cmd[1] = 0x04;
                cmd[2] = 0x00;

                cmd[3] = offset;
                cmd[4] = nBytes;
                cmd[5] = b1;
                cmd[6] = b2;

                byte bChkSum = 0;
                for (int i = 0; i < cmd.Length - 1; i++)
                    bChkSum += cmd[i];
                cmd[7] = bChkSum;

                // Your unmanaged pointer.

                // Allocate some unmanaged memory for those bytes.
                IntPtr pUnmanagedBytes = Marshal.AllocCoTaskMem(8);
                // Copy the managed byte array into the unmanaged array.
                Marshal.Copy(cmd, 0, pUnmanagedBytes, 8);

                bRet = _printLabel.SendBytesToPrinter(_printLabel.Printer, pUnmanagedBytes, 8);
            }
            catch (Exception)
            {
                bRet = false;
            }
            return bRet;


        }
    }
}
